/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.pinot.thirdeye.detection.algorithm;

import com.google.common.base.Preconditions;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import org.apache.commons.collections.MapUtils;
import org.apache.pinot.thirdeye.common.dimension.DimensionMap;
import org.apache.pinot.thirdeye.datalayer.dto.DetectionConfigDTO;
import org.apache.pinot.thirdeye.datalayer.dto.MergedAnomalyResultDTO;
import org.apache.pinot.thirdeye.detection.ConfigUtils;
import org.apache.pinot.thirdeye.detection.DataProvider;
import org.apache.pinot.thirdeye.detection.DetectionPipeline;
import org.apache.pinot.thirdeye.detection.DetectionPipelineResult;
import org.apache.pinot.thirdeye.detection.DetectionUtils;
import org.apache.pinot.thirdeye.detection.spi.model.AnomalySlice;


/**
 * The Merge wrapper. Supports time-based merging of anomalies along the same set of dimensions.
 * Forward scan only, greedy clustering. Does not merge clusters that converge.
 */
public class MergeWrapper extends DetectionPipeline {
  private static final String PROP_NESTED = "nested";
  private static final String PROP_CLASS_NAME = "className";
  private static final String PROP_MERGE_KEY = "mergeKey";
  private static final String PROP_DETECTOR_COMPONENT_NAME = "detectorComponentName";

  protected static final Comparator<MergedAnomalyResultDTO> COMPARATOR = new Comparator<MergedAnomalyResultDTO>() {
    @Override
    public int compare(MergedAnomalyResultDTO o1, MergedAnomalyResultDTO o2) {
      // earlier
      int res = Long.compare(o1.getStartTime(), o2.getStartTime());
      if (res != 0) return res;

      // pre-existing
      if (o1.getId() == null && o2.getId() != null) return 1;
      if (o1.getId() != null && o2.getId() == null) return -1;

      // more children
      return -1 * Integer.compare(o1.getChildren().size(), o2.getChildren().size());
    }
  };

  protected final List<Map<String, Object>> nestedProperties;
  protected final long maxGap; // max time gap for merge
  protected final long maxDuration; // max overall duration of merged anomaly
  protected final AnomalySlice slice;

  /**
   * Instantiates a new merge wrapper.
   *
   * @param provider the provider
   * @param config the config
   * @param startTime the start time
   * @param endTime the end time
   */
  public MergeWrapper(DataProvider provider, DetectionConfigDTO config, long startTime, long endTime) {
    super(provider, config, startTime, endTime);

    this.maxGap = MapUtils.getLongValue(config.getProperties(), "maxGap", 0);
    this.maxDuration = MapUtils.getLongValue(config.getProperties(), "maxDuration", Long.MAX_VALUE);
    this.slice = new AnomalySlice().withStart(startTime).withEnd(endTime);
    this.nestedProperties = new ArrayList<>();
    List<Map<String, Object>> nested = ConfigUtils.getList(config.getProperties().get(PROP_NESTED));
    for (Map<String, Object> properties : nested) {
      this.nestedProperties.add(new HashMap<>(properties));
    }
  }

  @Override
  public DetectionPipelineResult run() throws Exception {
    Map<String, Object> diagnostics = new HashMap<>();

    // generate anomalies
    List<MergedAnomalyResultDTO> generated = new ArrayList<>();

    int i = 0;
    Set<Long> lastTimeStamps = new HashSet<>();
    for (Map<String, Object> properties : this.nestedProperties) {
      DetectionConfigDTO nestedConfig = new DetectionConfigDTO();

      Preconditions.checkArgument(properties.containsKey(PROP_CLASS_NAME), "Nested missing " + PROP_CLASS_NAME);

      nestedConfig.setId(this.config.getId());
      nestedConfig.setName(this.config.getName());
      nestedConfig.setDescription(this.config.getDescription());
      nestedConfig.setProperties(properties);
      nestedConfig.setComponents(this.config.getComponents());
      DetectionPipeline pipeline = this.provider.loadPipeline(nestedConfig, this.startTime, this.endTime);

      DetectionPipelineResult intermediate = pipeline.run();
      lastTimeStamps.add(intermediate.getLastTimestamp());

      generated.addAll(intermediate.getAnomalies());
      diagnostics.put(String.valueOf(i), intermediate.getDiagnostics());

      i++;
    }

    // merge
    Set<MergedAnomalyResultDTO> all = new HashSet<>();
    all.addAll(retrieveAnomaliesFromDatabase(generated));
    all.addAll(generated);

    return new DetectionPipelineResult(this.merge(all), DetectionUtils.consolidateNestedLastTimeStamps(lastTimeStamps)).setDiagnostics(diagnostics);
  }

  protected List<MergedAnomalyResultDTO> retrieveAnomaliesFromDatabase(List<MergedAnomalyResultDTO> generated) {
    AnomalySlice effectiveSlice = this.slice
        .withStart(this.getStartTime(generated) - this.maxGap - 1)
        .withEnd(this.getEndTime(generated) + this.maxGap + 1);

    return new ArrayList<>(this.provider.fetchAnomalies(Collections.singleton(effectiveSlice), this.config.getId()).get(effectiveSlice));
  }

  // logic to do time-based merging.
  protected List<MergedAnomalyResultDTO> merge(Collection<MergedAnomalyResultDTO> anomalies) {
    List<MergedAnomalyResultDTO> input = new ArrayList<>(anomalies);
    Collections.sort(input, COMPARATOR);

    List<MergedAnomalyResultDTO> output = new ArrayList<>();

    Map<AnomalyKey, MergedAnomalyResultDTO> parents = new HashMap<>();
    for (MergedAnomalyResultDTO anomaly : input) {
      if (anomaly.isChild()) {
        continue;
      }

      AnomalyKey key = AnomalyKey.from(anomaly);
      MergedAnomalyResultDTO parent = parents.get(key);

      if (parent == null || anomaly.getStartTime() - parent.getEndTime() > this.maxGap) {
        // no parent, too far away
        parents.put(key, anomaly);
        output.add(anomaly);

      } else if (anomaly.getEndTime() <= parent.getEndTime() || anomaly.getEndTime() - parent.getStartTime() <= this.maxDuration) {
        // fully merge into existing
        parent.setEndTime(Math.max(parent.getEndTime(), anomaly.getEndTime()));

      } else if (parent.getEndTime() >= anomaly.getStartTime()) {
        // partially merge, truncate new
        long truncationTimestamp = Math.max(parent.getEndTime(), parent.getStartTime() + this.maxDuration);

        parent.setEndTime(truncationTimestamp);
        anomaly.setStartTime(Math.max(truncationTimestamp, anomaly.getStartTime()));
        anomaly.setEndTime(Math.max(truncationTimestamp, anomaly.getEndTime()));

        parents.put(key, anomaly);
        output.add(anomaly);

      } else {
        // default to new parent if merge not possible
        parents.put(key, anomaly);
        output.add(anomaly);

      }
    }

    return output;
  }

  protected long getStartTime(Iterable<MergedAnomalyResultDTO> anomalies) {
    long time = this.startTime;
    for (MergedAnomalyResultDTO anomaly : anomalies) {
      time = Math.min(anomaly.getStartTime(), time);
    }
    return time;
  }

  protected long getEndTime(Iterable<MergedAnomalyResultDTO> anomalies) {
    long time = this.endTime;
    for (MergedAnomalyResultDTO anomaly : anomalies) {
      time = Math.max(anomaly.getEndTime(), time);
    }
    return time;
  }

  protected static class AnomalyKey {
    final String metric;
    final String collection;
    final DimensionMap dimensions;
    final String mergeKey;
    final String componentKey;

    public AnomalyKey(String metric, String collection, DimensionMap dimensions, String mergeKey, String componentKey) {
      this.metric = metric;
      this.collection = collection;
      this.dimensions = dimensions;
      this.mergeKey = mergeKey;
      this.componentKey = componentKey;
    }

    public static AnomalyKey from(MergedAnomalyResultDTO anomaly) {
      return new AnomalyKey(anomaly.getMetric(), anomaly.getCollection(), anomaly.getDimensions(), anomaly.getProperties().get(PROP_MERGE_KEY),
          anomaly.getProperties().get(PROP_DETECTOR_COMPONENT_NAME));
    }

    @Override
    public boolean equals(Object o) {
      if (this == o) {
        return true;
      }
      if (!(o instanceof AnomalyKey)) {
        return false;
      }
      AnomalyKey that = (AnomalyKey) o;
      return Objects.equals(metric, that.metric) && Objects.equals(collection, that.collection) && Objects.equals(
          dimensions, that.dimensions) && Objects.equals(mergeKey, that.mergeKey) && Objects.equals(componentKey,
          that.componentKey);
    }

    @Override
    public int hashCode() {
      return Objects.hash(metric, collection, dimensions, mergeKey, componentKey);
    }
  }
}
